/*
 * Copyright (C) 2022 Nagisa Sekiguchi
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef ARSH_TOOLS_HIGHLIGHTER_FACTORY_H
#define ARSH_TOOLS_HIGHLIGHTER_FACTORY_H

#include "formatter.h"

namespace arsh::highlighter {

enum class FormatterType : unsigned char {
  NULL_,
  TERM_TRUECOLOR,
  TERM_256,
  HTML,
};

class FormatterFactory {
private:
  std::reference_wrapper<const StyleMap> styleMap;

  StrRefMap<FormatterType> supportedFormats; // name to actual formatter type

  StringRef formatName{"ansi"};

  std::string styleName{DEFAULT_STYLE_NAME};

  StringRef lineno;

  bool htmlFull{false};

  bool htmlTable{false};

  std::vector<StringRef> customStyles;

public:
  static constexpr const char *DEFAULT_STYLE_NAME = "darcula";

  explicit FormatterFactory(const StyleMap &map);

  const auto &getStyleMap() const { return this->styleMap.get(); }

  const auto &getSupportedFormats() const { return this->supportedFormats; }

  void setFormatName(StringRef name) { this->formatName = name; }

  void setStyleName(StringRef name) { this->styleName = name.toString(); }

  void setLineno(StringRef num) { this->lineno = num; }

  void setHTMLFull(bool set) { this->htmlFull = set; }

  void setHTMLTable(bool set) { this->htmlTable = set; }

  void setCustomStyles(std::vector<StringRef> &&values) { this->customStyles = std::move(values); }

  Result<std::unique_ptr<Formatter>, std::string> create(std::ostream &stream) const;
};

} // namespace arsh::highlighter

#endif // ARSH_TOOLS_HIGHLIGHTER_FACTORY_H
